import asyncio
from concurrent.futures import Future, ThreadPoolExecutor
from typing import AsyncIterator, Generator, Iterator, TypeVar

from .asyncio import run_in_executor, run_in_threadpool
from .utils import DummyStopIteration, next_without_stop_iteration

T = TypeVar("T")


def preactivate(iterator: Iterator[T]) -> Iterator[T]:
    iterator = iter(iterator)
    prefetched = next(iterator)
    return _preactivate(prefetched, iterator)


def _preactivate(prefetched: T, iterator: Iterator[T]) -> Iterator[T]:
    yield prefetched
    yield from iterator


async def apreactivate(aiterator: AsyncIterator[T]) -> AsyncIterator[T]:
    aiterator = aiter(aiterator)
    prefetched = await anext(aiterator)
    return _apreactivate(prefetched, aiterator)


async def _apreactivate(prefetched: T, aiterator: AsyncIterator[T]) -> AsyncIterator[T]:
    yield prefetched
    async for it in aiterator:
        yield it


def aprefetch_iterator(iterator: Iterator[T], p: ThreadPoolExecutor | None = None):
    iterator = iter(iterator)
    prefetched = run_in_executor(p, next_without_stop_iteration, iterator)

    return _aprefetch_iterator(iterator, p, prefetched)


async def _aprefetch_iterator(
    iterator: Iterator[T],
    p: ThreadPoolExecutor,
    prefetched: asyncio.Future[T],
):
    try:
        while True:
            try:
                ret = await prefetched
            except DummyStopIteration:
                break
            prefetched = run_in_executor(p, next_without_stop_iteration, iterator)
            yield ret
    finally:
        if p is not None:
            await run_in_threadpool(p.shutdown)


def prefetch_iterator(iterator: Iterator[T], p: ThreadPoolExecutor | None = None):
    if p is None:
        p = ThreadPoolExecutor(1)

    iterator = iter(iterator)
    prefetched = p.submit(next, iterator)

    return _prefetch_iterator(iterator, p, prefetched)


def _prefetch_iterator(
    iterator: Iterator[T],
    p: ThreadPoolExecutor,
    prefetched: Future[T],
) -> Generator[T, None, None]:
    with p:
        while True:
            try:
                rets = prefetched.result()
            except StopIteration:
                break
            prefetched = p.submit(next, iterator)
            yield rets


def aprefetch_aiterator(iterator: AsyncIterator[T], preactivate: bool = False):
    iterator = aiter(iterator)
    if preactivate:
        prefetched = asyncio.create_task(anext(iterator))
    else:
        prefetched = None

    return _aprefetch_aiterator(iterator, prefetched)


class __NULL:
    pass


async def _aprefetch_aiterator(
    iterator: Iterator[T],
    prefetched: asyncio.Task[T] | None,
):
    ret = __NULL
    while True:
        if prefetched is not None:
            try:
                ret = await asyncio.shield(prefetched)
            except StopAsyncIteration:
                break

        prefetched = asyncio.create_task(anext(iterator))
        if ret is not __NULL:
            yield ret


class BasePrefetcher:
    def __init__(self):
        self._p = ThreadPoolExecutor(1)
        self._future = None


class AsyncPrefetcher(BasePrefetcher):
    async def prefetch(self, func, *args, **kwargs):
        ret = None
        if self._future is not None:
            ret = await self._future
            self._future = None

        if func is None:
            return ret

        self._future = run_in_executor(self._p, func, *args, **kwargs)
        return ret

    async def __aenter__(self):
        self._p.__enter__()
        return self

    async def __aexit__(self, *args):
        await self.prefetch(None)
        self._p.__exit__(*args)


class Prefetcher(BasePrefetcher):
    def prefetch(self, func, *args, **kwargs):
        ret = None
        if self._future is not None:
            ret = self._future.result()
            self._future = None

        if func is None:
            return ret

        self._future = self._p.submit(func, *args, **kwargs)
        return ret

    def __enter__(self):
        self._p.__enter__()
        return self

    def __exit__(self, *args):
        self.prefetch(None)
        self._p.__exit__(*args)



import fractions
from abc import ABC, abstractmethod
from collections import defaultdict, deque
from typing import Callable, Iterable

import av
import av.filter
import av.filter.context
import numpy as np


class FilterContextOutput:
    def __init__(
        self, filter_context: av.filter.context.FilterContext, output_idx: int
    ):
        self._filter_context = filter_context
        self._output_idx = output_idx

    def push(self, frame: av.VideoFrame | None):
        self._filter_context.push(frame)

    def pull(self):
        return self._filter_context.pull()

    def link_to(self, input: "FilterContext", input_idx: int = 0):
        self._filter_context.link_to(
            input._filter_context,
            output_idx=self._output_idx,
            input_idx=input_idx,
        )

    def __rshift__(self, right: "FilterContext"):
        self.link_to(right, input_idx=0)
        return right.output()


class FilterContext:
    def __init__(self, filter_context: av.filter.context.FilterContext):
        self._filter_context = filter_context

    def process_command(
        self, cmd: str, arg: str | None = None, res_len: int = 1024, flags: int = 0
    ):
        self._filter_context.process_command(
            cmd=cmd, arg=arg, res_len=res_len, flags=flags
        )

    def output(self) -> list[FilterContextOutput] | FilterContextOutput:
        len_output = len(self._filter_context.outputs)
        if len_output < 2:
            return FilterContextOutput(self._filter_context, 0)
        else:
            return [
                FilterContextOutput(self._filter_context, i) for i in range(len_output)
            ]

    def __rrshift__(self, left: Iterable[FilterContextOutput] | FilterContextOutput):
        if isinstance(left, Iterable):
            for i, it in enumerate(left):
                it.link_to(self, input_idx=i)
        elif isinstance(left, FilterContext):
            left >> self
        else:
            raise NotImplementedError

        return self.output()


class Graph:
    def __init__(self):
        self._graph = av.filter.Graph()

    def add(
        self, filter: str | av.filter.Filter, args: str = None, **kwargs: str
    ) -> FilterContext:
        return FilterContext(self._graph.add(filter, args, **kwargs))

    def add_buffer(
        self,
        template: av.VideoStream | None = None,
        width: int | None = None,
        height: int | None = None,
        format: av.VideoFormat | None = None,
        name: str | None = None,
        time_base: fractions.Fraction | None = None,
    ):
        buffer = self._graph.add_buffer(
            template=template,
            width=width,
            height=height,
            format=format,
            name=name,
            time_base=time_base,
        )
        return FilterContextOutput(buffer, 0)

    def push(self, frame: av.VideoFrame | None):
        self._graph.push(frame)

    def pull(self):
        return self._graph.pull()

    def configure(self, auto_buffer: bool = True, force: bool = False):
        return self._graph.configure(auto_buffer=auto_buffer, force=force)


class Filter(ABC):
    _graph: Graph
    _buffers: list[FilterContextOutput]

    def __init__(
        self,
        graph: Graph,
        buffers: list[FilterContextOutput],
    ):
        self._graph = graph
        self._buffers = buffers

    def __call__(
        self, *frames: av.VideoFrame, **commands: dict[str, str]
    ) -> av.VideoFrame:
        self.process_command(**commands)
        self.push(*frames)
        ret = self._graph.pull()
        ret.pts = None
        return ret

    def process_command(self, **commands):
        "process command"

    def push(self, *frames):
        for frame, buffer in zip(frames, self._buffers):
            buffer.push(frame)

    def close(self):
        if self._graph is not None:
            self._graph.push(None)


class ScaledOverlay(Filter):
    def __init__(
        self,
        background_width: int,
        background_height: int,
        width: int,
        height: int,
        pix_fmt: str = "rgba",
    ):
        graph = Graph()
        super().__init__(
            graph,
            [
                self._graph.add_buffer(
                    width=background_width, height=background_height, format=pix_fmt
                ),
                self._graph.add_buffer(width=width, height=height, format=pix_fmt),
            ],
        )

        self._scale1 = self._graph.add("scale")
        self._scale2 = self._graph.add("scale")
        self._overlay = self._graph.add("overlay", format=pix_fmt.replace("a", ""))

        (
            (
                (self._buffers[0] >> self._scale1),
                (self._buffers[1] >> self._scale2),
            )
            >> self._overlay
            >> self._graph.add("buffersink")
        )

        self._graph.configure()

    def process_command(
        self,
        x: int,
        y: int,
        background_width: int = -1,
        background_height: int = -1,
        width: int = -1,
        height: int = -1,
    ):
        self._scale1.process_command("width", f"{background_width}")
        self._scale1.process_command("height", f"{background_height}")
        self._scale2.process_command("width", f"{width}")
        self._scale2.process_command("height", f"{height}")
        self._overlay.process_command("x", f"{x}")
        self._overlay.process_command("y", f"{y}")


class RebuildableFilter(Filter):
    def __init__(self):
        super().__init__(None, [])
        self._kwargs = {}

    def rebuild_graph(self, **kwargs):
        if self._kwargs != kwargs:
            self.close()
            self._graph = Graph()
            self._rebuild_graph(**kwargs)
            self._graph.configure()
            self._kwargs = kwargs

    @abstractmethod
    def _rebuild_graph(**kwargs):
        "rebuild graph impl"


def to_rgba(reader: Iterable[av.VideoFrame]):
    reader = iter(reader)
    first_frame = next(reader)

    graph = Graph()
    buffer = graph.add_buffer(template=first_frame)
    buffer >> graph.add("format", pix_fmts="rgba") >> graph.add("buffersink")
    graph.configure()

    graph.push(first_frame)
    yield graph.pull()

    for frame in reader:
        graph.push(frame)
        yield graph.pull()

    graph.push(None)

def to_array(iterator: list[av.VideoFrame]):
    for frame in iterator:
        yield frame.to_ndarray()


class _AlphaExtractor(RebuildableFilter):
    def __init__(self):
        super().__init__()

    def _rebuild_graph(self, width: int, height: int, pix_fmt: str):
        self._buffers = [
            self._graph.add_buffer(
                width=width,
                height=height,
                format=pix_fmt,
                time_base=fractions.Fraction(1, 1000),
            ),
        ]

        (
            self._buffers
            >> self._graph.add("alphaextract")
            >> self._graph.add("buffersink")
        )

    def __call__(self, frame: av.VideoFrame):
        self.rebuild_graph(
            width=frame.width, height=frame.height, pix_fmt=frame.format.name
        )
        return super().__call__(frame)


class AlphaExtractor:
    def __init__(self):
        self.alpha_extractors = defaultdict(_AlphaExtractor)

    def __call__(self, frame: av.VideoFrame):
        assert frame.height % 2 == 0

        return self.alpha_extractors[frame.format.name](frame)

    def close(self):
        for it in self.alpha_extractors.values():
            it.close()


class BaseAlphaMerger:
    @abstractmethod
    def push_image(self, frame: av.VideoFrame):
        """push image to merger"""

    @abstractmethod
    def push_alpha(self, frame: av.VideoFrame):
        """push alpha to merger"""

    @abstractmethod
    def pull(self):
        """pull merged image"""


class AlphaMerger(BaseAlphaMerger):
    def __init__(self, image_pix_fmt: str, alpha_pix_fmt: str):
        self.alpha_mergers = defaultdict(
            lambda: _AlphaMerger(image_pix_fmt, alpha_pix_fmt)
        )

    def push_image(self, frame: av.VideoFrame):
        self.alpha_mergers[(frame.width, frame.height)].push_image(frame)

    def push_alpha(self, frame: av.VideoFrame):
        self.alpha_mergers[(frame.width, frame.height)].push_alpha(frame)

    def pull(self) -> av.VideoFrame:
        keys = list(self.alpha_mergers)
        for key in keys:
            ret = self.alpha_mergers[key].pull()
            if ret is not None:
                return ret

            if len(keys) > 1:
                del self.alpha_mergers[key]

    def close(self):
        for it in self.alpha_mergers.values():
            it.close()


class _AlphaMerger(RebuildableFilter, BaseAlphaMerger):
    def __init__(self, image_pix_fmt: str, alpha_pix_fmt: str):
        super().__init__()
        self.image_pix_fmt = image_pix_fmt
        self.alpha_pix_fmt = alpha_pix_fmt

    def _rebuild_graph(
        self,
        width: int,
        height: int,
    ):
        self._buffers = [
            self._graph.add_buffer(
                width=width,
                height=height,
                format=self.image_pix_fmt,
                time_base=fractions.Fraction(1, 1000),
            ),
            self._graph.add_buffer(
                width=width,
                height=height,
                format=self.alpha_pix_fmt,
                time_base=fractions.Fraction(1, 1000),
            ),
        ]

        (
            (self._buffers[0], self._buffers[1] >> self._graph.add("format", "gray"))
            >> self._graph.add("alphamerge")
            >> self._graph.add("buffersink")
        )

    def push_image(self, frame: av.VideoFrame):
        self.rebuild_graph(width=frame.width, height=frame.height)
        self._buffers[0].push(frame)

    def push_alpha(self, frame: av.VideoFrame):
        self.rebuild_graph(width=frame.width, height=frame.height)
        self._buffers[1].push(frame)

    def pull(self) -> av.VideoFrame:
        try:
            return self._graph.pull()
        except BlockingIOError:
            return None


class NotAlphaMerger(BaseAlphaMerger):
    def __init__(self):
        self.queue = deque()

    def push_image(self, frame: av.VideoFrame):
        self.queue.append(frame)

    def push_alpha(self, frame: av.VideoFrame):
        raise NotImplementedError

    def pull(self) -> av.VideoFrame:
        try:
            return self.queue.popleft()
        except IndexError:
            return None


def get_dst_size(dst_size: tuple[int, int], background_image: np.ndarray):
    target_height, target_width = background_image.shape[:2]

    width, height = dst_size
    if target_height / height < target_width / width:
        width = round(target_height / height * width)
        height = target_height
    else:
        height = round(target_width / width * height)
        width = target_width

    width, height = width - width % 16, height - height % 16

    bg_top = (target_height - height) // 2
    bg_left = (target_width - width) // 2

    return (width, height), background_image[
        bg_top : bg_top + height, bg_left : bg_left + width, :
    ]


def get_src_size(
    left: float,
    top: float,
    height: float,
    dst_size: tuple[int, int],
    src_size: tuple[int, int],
):
    dst_width, dst_height = dst_size
    src_width, src_height = src_size

    target_frame_height = dst_height * height
    frame_width = min(
        round(src_width * target_frame_height / src_height),
        dst_width,
    )
    frame_height = round(src_height * frame_width / src_width)

    left = (left + 1) / 2
    left_limit = dst_width - frame_width

    x = round(left * left_limit)
    y = round(top * dst_height)

    return (x, y), (frame_width, frame_height)


class ScaledCenterCropOverlay(Filter):
    def __init__(
        self,
        background_width: int,
        background_height: int,
        width: int,
        height: int,
        pix_fmt: str = "rgba",
    ):
        graph = Graph()
        super().__init__(
            graph,
            [
                graph.add_buffer(
                    width=background_width, height=background_height, format=pix_fmt
                ),
                graph.add_buffer(width=width, height=height, format=pix_fmt),
            ],
        )

        self._scale1 = self._graph.add(
            "scale", force_original_aspect_ratio="increase", flags="fast_bilinear"
        )
        self._scale2 = self._graph.add("scale", flags="fast_bilinear")
        self._crop = self._graph.add("crop", x="(iw-ow)/2", y="(ih-oh)/2")

        self._overlay = self._graph.add("overlay", format=pix_fmt.replace("a", ""))

        (
            (
                (self._buffers[0] >> self._scale1 >> self._crop),
                (self._buffers[1] >> self._scale2),
            )
            >> self._overlay
            >> self._graph.add("buffersink")
        )

        self._graph.configure()

    def process_command(
        self,
        x: int,
        y: int,
        background_width: int = -1,
        background_height: int = -1,
        width: int = -1,
        height: int = -1,
    ):
        self._scale1.process_command("w", f"{background_width}")
        self._scale1.process_command("h", f"{background_height}")
        self._crop.process_command("w", f"{background_width}")
        self._crop.process_command("h", f"{background_height}")
        self._scale2.process_command("w", f"{width}")
        self._scale2.process_command("h", f"{height}")
        self._overlay.process_command("x", f"{x}")
        self._overlay.process_command("y", f"{y}")


class Paste(Filter):
    def __init__(
        self,
        background_width: int,
        background_height: int,
        width: int,
        height: int,
        pix_fmt: str = "rgba",
    ):
        graph = Graph()
        super().__init__(
            graph,
            [
                graph.add_buffer(
                    width=background_width, height=background_height, format=pix_fmt
                ),
                graph.add_buffer(width=width, height=height, format=pix_fmt),
            ],
        )

        self._scale = self._graph.add("scale")
        self._overlay1 = self._graph.add(
            "overlay", format=pix_fmt.replace("a", "").replace("p", "")
        )
        self._overlay2 = self._graph.add(
            "overlay", format=pix_fmt.replace("a", "").replace("p", "")
        )

        buffer11, buffer12 = self._buffers[0] >> graph.add("split")
        buffer21, buffer22 = (
            self._buffers[1]
            >> self._scale
            >> graph.add("format", pix_fmt)
            >> graph.add("split")
        )

        (
            (
                (
                    buffer11 >> graph.add("alphaextract"),
                    buffer21 >> graph.add("alphaextract"),
                )
                >> self._overlay1
                >> graph.add("format", "gray"),
                (
                    buffer12 >> graph.add("format", "rgb24"),
                    buffer22 >> graph.add("format", "rgb24"),
                )
                >> self._overlay2,
            )
            >> graph.add("alphamerge")
            >> graph.add("buffersink")
        )

        self._graph.configure()

    def process_command(
        self,
        x: int,
        y: int,
        w: int,
        h: int,
    ):
        self._scale.process_command("w", f"{w}")
        self._scale.process_command("h", f"{h}")
        self._overlay1.process_command("x", f"{x}")
        self._overlay1.process_command("y", f"{y}")
        self._overlay2.process_command("x", f"{x}")
        self._overlay2.process_command("y", f"{y}")



import fractions
import threading
from abc import abstractmethod
from collections import defaultdict, deque
from typing import Callable, Iterable

import av
import av.filter
import av.filter.context
import numpy as np

from .codecs import AsyncDecoder, AsyncEncoder, VideoEncoder
from .filters import FilterContextOutput, Graph
from .io import BasePyAVReader, PyAVReader, PyAVWriter

type PyAVDisposableReader = PyAVReader



import asyncio
from typing import AsyncIterable

import av

from ..asyncio import run_in_threadpool
from ..io import PopIO
from ..prefetch import aprefetch_iterator
from .io import PyAVReader, PyAVWriter


class AsyncDecoder:
    def __init__(self, aiterator: AsyncIterable[bytes], **kwargs):
        self._aiterator = aiterator
        self._f = PopIO()
        self._kwargs = kwargs

    def decode(self) -> AsyncIterable[av.VideoFrame | av.AudioFrame]:
        async def _pull():
            try:
                async for it in self._aiterator:
                    self._f.write(it)
            finally:
                self._f.close()

        pull_task = asyncio.create_task(_pull())

        def _decode():
            try:
                yield from PyAVReader(self._f, **self._kwargs)
            finally:
                if pull_task.done():
                    pull_task.result()
                else:
                    pull_task.cancel()

        return aprefetch_iterator(_decode())


class AsyncEncoder:
    def __init__(self, writer: PyAVWriter):
        self._writer = writer
        self._f = PopIO()
        writer.lazy_register_path(self._f)

    async def encode(self, frame: av.VideoFrame | av.AudioFrame):
        if isinstance(frame, av.VideoFrame):
            await run_in_threadpool(self._writer.write_video_frame_lazy, frame)
        elif isinstance(frame, av.AudioFrame):
            await run_in_threadpool(self._writer.write_audio_frame_lazy, frame)
        else:
            raise NotImplementedError

    async def aclose(self):
        await run_in_threadpool(self._writer.__exit__, None, None, None)
        await run_in_threadpool(self._f.close)

    def __aiter__(self):
        return self

    async def __anext__(self):
        if ret := await run_in_threadpool(self._f.read):
            return ret
        else:
            raise StopAsyncIteration


class VideoEncoder:
    def __init__(
        self, *, stream: av.VideoStream, codec_context: av.VideoCodecContext, **kwargs
    ):
        for key, value in kwargs.items():
            # if value is not None:
            setattr(codec_context, key, value)

        self.stream = stream
        self.codec_context = codec_context

    @classmethod
    def from_stream(cls, stream: av.VideoStream, width: int, height: int, **kwargs):
        stream_cc = stream.codec_context
        cc = stream.codec.create("video")

        _kwargs = {
            "pix_fmt": stream_cc.pix_fmt,
            "bit_rate": stream_cc.bit_rate,
            "time_base": stream_cc.time_base,
            "color_primaries": stream_cc.color_primaries,
            "color_range": stream_cc.color_range,
            "color_trc": stream_cc.color_trc,
            "framerate": stream_cc.framerate,
            "gop_size": stream_cc.gop_size,
            "qmax": stream_cc.qmax,
            "qmin": stream_cc.qmin,
            "options": stream_cc.options,
        }

        for key, value in kwargs.items():
            _kwargs[key] = value

        return cls(
            stream=stream,
            codec_context=cc,
            width=width,
            height=height,
            **_kwargs,
        )

    def encode(self, frame: av.VideoFrame):
        packets = self.codec_context.encode(frame)
        for packet in packets:
            packet.stream = self.stream
        return packets



import asyncio
import time
from typing import AsyncGenerator, AsyncIterable, Generic, TypeVar

from .asyncio import run_in_threadpool
from .utils import DummyStopIteration, next_without_stop_iteration


async def to_aiterator(iterator):
    iterator = await run_in_threadpool(iter, iterator)
    while True:
        try:
            yield await run_in_threadpool(next_without_stop_iteration, iterator)
        except DummyStopIteration:
            break


async def rate_limit_iterator(aiterator, iters_per_second):
    start = time.time()
    i = 0
    async for it in aiterator:
        yield it
        await asyncio.sleep((i / iters_per_second) - (time.time() - start))
        i += 1


T = TypeVar("T")


class BaseStreamQueue(Generic[T]):
    _CLOSE = object()

    def __init__(self):
        self.queue = asyncio.Queue()

    async def close(self):
        await self.queue.put(self._CLOSE)


class StreamQueue(BaseStreamQueue):
    queue: asyncio.Queue[T]

    def put(self, data: T) -> None:
        return self.queue.put(data)

    async def __aiter__(self) -> AsyncGenerator[T]:
        while (it := await self.queue.get()) is not self._CLOSE:
            yield it


class StreamIteratorQueue(BaseStreamQueue):
    queue: asyncio.Queue[AsyncIterable[T]]

    def put(self, data: AsyncIterable[T]) -> None:
        return self.queue.put(data)

    async def __aiter__(self) -> AsyncGenerator[T]:
        while (iterator := await self.queue.get()) is not self._CLOSE:
            async for it in iterator:
                yield it



import av
import numba

@numba.jit(
    (
        numba.types.MemoryView(numba.uint8, 1, "C"),
        numba.types.MemoryView(numba.uint8, 1, "C"),
        numba.int64,
        numba.int64,
        numba.int64,
        numba.int64,
        numba.int64,
        numba.int64,
        numba.int64,
        numba.int64,
    ),
    nopython=True,
    nogil=True,
)
def _frame_paste_rgba(
    dst_view: memoryview,
    src_view: memoryview,
    x: int,
    y: int,
    dw: int,
    dh: int,
    sw: int,
    sh: int,
    src_stride: int,
    dst_stride: int,
):
    sx = sy = 0

    if x < 0:
        sx -= x
        sw += x
        x = 0
    if y < 0:
        sy -= y
        sh += y
        y = 0

    sw = min(sw, sw - sx, dw - x)
    sh = min(sh, sh - sy, dh - y)

    if sw <= 0 or sh <= 0:
        return

    s_base = sy * src_stride + sx * 4
    d_base = y * dst_stride + x * 4
    row_bytes = sw * 4

    for r in range(sh):
        so = s_base + r * src_stride
        do = d_base + r * dst_stride
        dst_view[do : do + row_bytes] = src_view[so : so + row_bytes]


def frame_paste_rgba(dst: av.VideoFrame, src: av.VideoFrame, x: int, y: int):
    assert src.format.name == "rgba" and dst.format.name == "rgba"

    dst.make_writable()

    src_plane = src.planes[0]
    dst_plane = dst.planes[0]

    _frame_paste_rgba(
        memoryview(dst_plane),
        memoryview(src_plane),
        x,
        y,
        dst.width,
        dst.height,
        src.width,
        src.height,
        src_plane.line_size,
        dst_plane.line_size,
    )
    return dst


@numba.jit(
    (
        numba.types.MemoryView(numba.uint8, 1, "C"),
        numba.types.MemoryView(numba.uint8, 1, "C"),
        numba.int64,
        numba.int64,
        numba.int64,
        numba.int64,
        numba.int64,
        numba.int64,
        numba.int64,
        numba.int64,
    ),
    nopython=True,
    nogil=True,
)
def _blit_8bit_planar(
    dst_view: memoryview,
    src_view: memoryview,
    dx: int,
    dy: int,
    sx: int,
    sy: int,
    w: int,
    h: int,
    dst_stride: int,
    src_stride: int,
):
    if w <= 0 or h <= 0:
        return
    row_bytes = w  # 1 byte per pixel in 8-bit planar
    s_base = sy * src_stride + sx
    d_base = dy * dst_stride + dx
    for r in range(h):
        so = s_base + r * src_stride
        do = d_base + r * dst_stride
        dst_view[do : do + row_bytes] = src_view[so : so + row_bytes]


@numba.jit(
    (
        numba.types.MemoryView(numba.uint8, 1, "C"),
        numba.types.MemoryView(numba.uint8, 1, "C"),
        numba.types.MemoryView(numba.uint8, 1, "C"),
        numba.types.MemoryView(numba.uint8, 1, "C"),
        numba.types.MemoryView(numba.uint8, 1, "C"),
        numba.types.MemoryView(numba.uint8, 1, "C"),
        numba.types.MemoryView(numba.uint8, 1, "C"),
        numba.types.MemoryView(numba.uint8, 1, "C"),
        numba.int64,
        numba.int64,
        numba.int64,
        numba.int64,
        numba.int64,
        numba.int64,
        numba.int64,
        numba.int64,
        numba.int64,
        numba.int64,
        numba.int64,
        numba.int64,
        numba.int64,
        numba.int64,
    ),
    nopython=True,
    nogil=True,
)
def _frame_paste_yuva420p(
    dst_view_y: memoryview,
    dst_view_u: memoryview,
    dst_view_v: memoryview,
    dst_view_a: memoryview,
    src_view_y: memoryview,
    src_view_u: memoryview,
    src_view_v: memoryview,
    src_view_a: memoryview,
    x: int,
    y: int,
    dw: int,
    dh: int,
    sw: int,
    sh: int,
    src_stride_y: int,
    src_stride_u: int,
    src_stride_v: int,
    src_stride_a: int,
    dst_stride_y: int,
    dst_stride_u: int,
    dst_stride_v: int,
    dst_stride_a: int,
):
    sw0, sh0 = sw, sh
    dw, dh = dw, dh

    sx = max(0, -x)
    sy = max(0, -y)
    dx = max(0, x)
    dy = max(0, y)

    w = min(sw0 - sx, dw - dx)
    h = min(sh0 - sy, dh - dy)

    if w <= 0 or h <= 0:
        return

    _blit_8bit_planar(
        dst_view_y,
        src_view_y,
        dx,
        dy,
        sx,
        sy,
        w,
        h,
        dst_stride_y,
        src_stride_y,
    )

    _blit_8bit_planar(
        dst_view_a,
        src_view_a,
        dx,
        dy,
        sx,
        sy,
        w,
        h,
        dst_stride_a,
        src_stride_a,
    )

    hsub = 1
    vsub = 1

    sx_c = sx >> hsub
    sy_c = sy >> vsub
    dx_c = dx >> hsub
    dy_c = dy >> vsub

    src_cw = (sw0 + (1 << hsub) - 1) >> hsub
    src_ch = (sh0 + (1 << vsub) - 1) >> vsub
    dst_cw = (dw + (1 << hsub) - 1) >> hsub
    dst_ch = (dh + (1 << vsub) - 1) >> vsub

    w_c = ((sx + w + (1 << hsub) - 1) >> hsub) - (sx >> hsub)
    h_c = ((sy + h + (1 << vsub) - 1) >> vsub) - (sy >> vsub)

    w_c = min(w_c, src_cw - sx_c, dst_cw - dx_c)
    h_c = min(h_c, src_ch - sy_c, dst_ch - dy_c)

    if w_c > 0 and h_c > 0:
        _blit_8bit_planar(
            dst_view_u,
            src_view_u,
            dx_c,
            dy_c,
            sx_c,
            sy_c,
            w_c,
            h_c,
            dst_stride_u,
            src_stride_u,
        )
        _blit_8bit_planar(
            dst_view_v,
            src_view_v,
            dx_c,
            dy_c,
            sx_c,
            sy_c,
            w_c,
            h_c,
            dst_stride_v,
            src_stride_v,
        )


def frame_paste_yuva420p(dst: av.VideoFrame, src: av.VideoFrame, x: int, y: int):
    assert src.format.name == "yuva420p" and dst.format.name == "yuva420p"

    dst.make_writable()

    src_Y, src_U, src_V, src_A = src.planes
    dst_Y, dst_U, dst_V, dst_A = dst.planes

    _frame_paste_yuva420p(
        memoryview(dst_Y),
        memoryview(dst_U),
        memoryview(dst_V),
        memoryview(dst_A),
        memoryview(src_Y),
        memoryview(src_U),
        memoryview(src_V),
        memoryview(src_A),
        x,
        y,
        dst.width,
        dst.height,
        src.width,
        src.height,
        src_Y.line_size,
        src_U.line_size,
        src_V.line_size,
        src_A.line_size,
        dst_Y.line_size,
        dst_U.line_size,
        dst_V.line_size,
        dst_A.line_size,
    )

    return dst



import fractions
import functools
import io
import threading
from abc import ABC, abstractmethod
from concurrent.futures import Future, ThreadPoolExecutor
from pathlib import Path
from types import TracebackType
from typing import Generator, Iterable

import av
import av.container
import av.frame
import av.stream
import numpy as np
from pydub import AudioSegment

from ..asyncio import run_in_threadpool
from .filters import AlphaExtractor, AlphaMerger, BaseAlphaMerger, NotAlphaMerger

TIME_BASE = fractions.Fraction(1, 90000)


class PyAVInterface(ABC):
    _container: av.container.Container
    _streams: tuple[av.VideoStream, ...]

    def __init__(self):
        self.__container_init_lock = threading.Lock()

    def _init_container(self):
        if self._container is None:
            with self.__container_init_lock:
                self._create_container()

    @abstractmethod
    def _create_container(self):
        """create container"""

    @property
    def container(self) -> av.container.Container:
        self._init_container()
        return self._container

    @property
    def streams(self) -> tuple[av.VideoStream, ...]:
        self._init_container()
        return self._streams

    @property
    def fps(self):
        return self.streams[0].base_rate or self.streams[0].codec_context.framerate

    @property
    def width(self):
        return self.streams[0].codec_context.width

    @property
    def height(self):
        return self.streams[0].codec_context.height

    @property
    def pix_fmt(self):
        return self.streams[0].format.name

    def __enter__(self):
        self.container.__enter__()
        return self

    def __exit__(self, *args):
        self.container.__exit__(*args)

    async def __aenter__(self):
        await run_in_threadpool(self.container.__enter__)

    async def __aexit__(self):
        await run_in_threadpool(self.container.__exit__)


class BasePyAVReader(PyAVInterface):
    container: av.container.InputContainer

    def __init__(
        self,
        path,
        *,
        format: str,
        buffer_size: int,
        filter: tuple[type[av.frame.Frame]],
        options={},
    ):
        super().__init__()
        self._container = None
        self._path = path
        self._format = format
        self._buffer_size = buffer_size
        self._filter = filter
        self._options = options

        self._codec_contexts = {}

    @abstractmethod
    def __iter__(self) -> Generator[av.VideoFrame | av.AudioFrame, None, None]:
        raise NotImplementedError

    def _create_container(self):
        if self._container is None:
            container = av.open(
                self._path,
                "r",
                format=self._format,
                buffer_size=self._buffer_size,
                options=self._options,
            )

            self._streams = tuple()
            if av.VideoFrame in self._filter:
                self._streams = container.streams.video
                for stream in self._streams:
                    if stream.codec_context.name in ("vp8", "vp9"):
                        if stream.codec_context.name == "vp8":
                            codec_name = "libvpx"
                        elif stream.codec_context.name == "vp9":
                            codec_name = "libvpx-vp9"
                        codec = av.codec.Codec(codec_name, "r")
                        self._codec_contexts[stream] = codec.create()
                    else:
                        self._codec_contexts[stream] = stream.codec_context

            self._audio_streams = tuple()
            if av.AudioFrame in self._filter:
                self._audio_streams = container.streams.audio
                for stream in self._audio_streams:
                    self._codec_contexts[stream] = stream.codec_context

            self._container = container

    @property
    def codec_contexts(self) -> dict[av.stream.Stream, av.CodecContext]:
        self._init_container()
        return self._codec_contexts

    @property
    def audio_streams(self) -> tuple[av.AudioStream, ...]:
        self._init_container()
        return self._audio_streams


class PyAVReader(BasePyAVReader):
    def __init__(
        self,
        path,
        start=0,
        end=(2 << 62) - 1,
        *,
        format=None,
        buffer_size=32768,
        filter=(av.VideoFrame, av.AudioFrame),
        options={},
    ):
        super().__init__(
            path, format=format, buffer_size=buffer_size, filter=filter, options=options
        )
        self.start = start
        self.end = end
        self._alpha_merger = None

    @property
    def alpha_merger(self) -> BaseAlphaMerger:
        if self._alpha_merger is None:
            if len(self.streams) < 2:
                self._alpha_merger = NotAlphaMerger()
            elif len(self.streams) == 2:
                self._alpha_merger = AlphaMerger(
                    self.streams[0].format.name, self.streams[1].format.name
                )
            else:
                raise NotImplementedError
        return self._alpha_merger

    def __iter__(self) -> Generator[av.VideoFrame | av.AudioFrame, None, None]:
        with self:
            for packet in self.container.demux(self.streams + self.audio_streams):
                for frame in self.codec_contexts[packet.stream].decode(packet):
                    if (
                        packet.stream in self.streams
                        and not (
                            self.start
                            <= round(frame.pts * self.fps * frame.time_base)
                            < self.end
                        )
                        or packet.stream in self.audio_streams
                        and not (
                            self.start - frame.time_base
                            <= frame.pts * frame.time_base
                            < self.end + frame.time_base
                        )
                    ):
                        continue

                    if packet.stream in self.audio_streams:
                        yield frame
                    elif packet.stream is self.streams[0]:
                        self.alpha_merger.push_image(frame)
                    else:
                        self.alpha_merger.push_alpha(frame)

                while (result := self.alpha_merger.pull()) is not None:
                    yield result
        if isinstance(self.alpha_merger, AlphaMerger):
            self.alpha_merger.close()


def create_stream(
    container: av.container.OutputContainer,
    codec_name: str,
    rate: int | fractions.Fraction,
    width: int,
    height: int,
    pix_fmt: str,
    bit_rate: int,
    time_base: fractions.Fraction,
    options: dict,
):
    stream = container.add_stream(
        codec_name=codec_name,
        rate=rate,
        width=width,
        height=height,
        pix_fmt=pix_fmt,
        bit_rate=bit_rate,
        time_base=time_base,
    )
    stream.options = options
    return stream


class PyAVWriter(PyAVInterface):
    container: av.container.OutputContainer

    def __init__(
        self,
        path: str | Path | io.IOBase | None,
        fps: fractions.Fraction = None,
        *,
        width: int = 640,
        height: int = 480,
        codec_name="libvpx-vp9",
        pix_fmt="yuva420p",
        buffer_size=32768,
        bit_rate=1024 * 1024,
        alpha_stream: bool | str = False,
        audio_codec_name=None,
        audio_sample_rate=48000,
        audio_format="s16",
        audio_layout="stereo",
        audio_bit_rate=192000,
        format=None,
        options={},
        container_options={},
    ):
        super().__init__()

        assert codec_name is not None or audio_codec_name is not None

        if codec_name is not None:
            if pix_fmt == "rgb24" and codec_name == "rawvideo" and alpha_stream:
                pix_fmt = "rgba"
                alpha_stream = False
            elif (
                pix_fmt == "yuv420p"
                and codec_name.startswith("libvpx")
                and alpha_stream
            ):
                pix_fmt = "yuva420p"
                alpha_stream = False
            elif (pix_fmt.startswith("yuva") or pix_fmt == "rgba") and alpha_stream:
                alpha_stream = False

        self._path = path
        self._width = width
        self._height = height
        self._fps = fps
        self._codex_contexts: dict[av.VideoStream, av.VideoCodecContext] = {}
        self._codec_name = codec_name
        self._pix_fmt = pix_fmt
        self._buffer_size = buffer_size
        self._bit_rate = bit_rate
        self._alpha_stream = alpha_stream
        self._audio_codec_name = audio_codec_name
        self._audio_sample_rate = audio_sample_rate
        self._audio_format = audio_format
        self._audio_layout = audio_layout
        self._audio_bit_rate = audio_bit_rate
        self._format = format
        self._options = options
        self._container_options = container_options

        self._container = None
        self._alpha_extractor = None

        self.__frames = 0

        self.pool = None
        self.future: Future[av.VideoFrame | av.AudioFrame] = None

        self.write_lazy = self.lazy(self.write)
        self.write_video_frame_lazy = self.lazy(self.write_video_frame)
        self.write_audio_lazy = self.lazy(self.write_audio)
        self.write_audio_frame_lazy = self.lazy(self.write_audio_frame)

    def lazy_register_path(self, path: str | Path | io.IOBase):
        if self._path is not None:
            raise ValueError
        self._path = path

    def _create_container(self):
        if self._path is None:
            raise ValueError

        container = av.open(
            self._path,
            "w",
            buffer_size=self._buffer_size,
            format=self._format,
            options=self._options,
            container_options=self._container_options,
        )
        streams = []
        if self._codec_name is not None:
            pix_fmts = [self._pix_fmt]
            if self._alpha_stream:
                pix_fmts.append(
                    self._pix_fmt if self._alpha_stream == True else self._alpha_stream
                )

            for pf in pix_fmts:
                stream = create_stream(
                    container,
                    codec_name=self._codec_name,
                    rate=self._fps,
                    width=self._width,
                    height=self._height,
                    pix_fmt=pf,
                    bit_rate=self._bit_rate,
                    time_base=TIME_BASE,
                    options=self._options,
                )
                streams.append(stream)

        audio_stream = None
        if self._audio_codec_name is not None:
            audio_stream = container.add_stream(
                codec_name=self._audio_codec_name, rate=self._audio_sample_rate
            )
            audio_stream.format = self._audio_format
            audio_stream.layout = self._audio_layout
            audio_stream.bit_rate = self._audio_bit_rate

        self._streams = streams
        self._audio_stream = audio_stream
        self._container = container

    @property
    def audio_stream(self) -> av.AudioStream:
        self._init_container()
        return self._audio_stream

    @property
    def alpha_extractor(self):
        if self._alpha_extractor is None and self._alpha_stream:
            self._alpha_extractor = AlphaExtractor()
        return self._alpha_extractor

    def array_to_frame(self, array):
        if self.streams[0].pix_fmt.startswith("yuva") or len(self.streams) == 2:
            frame = av.VideoFrame.from_ndarray(array, format="rgba")
        else:
            frame = av.VideoFrame.from_ndarray(array[..., :3], format="rgb24")
        return frame

    def lazy(self, func):
        @functools.wraps(func)
        def _func(*args, **kwargs):
            if self.pool is None:
                self.pool = ThreadPoolExecutor(1)

            if self.future is not None:
                self.future.result()
                del self.future
            self.future = self.pool.submit(func, *args, **kwargs)

        return _func

    def write(self, array):
        frame = self.array_to_frame(array)
        self.write_video_frame(frame)

    def create_codec_context(self, stream: av.VideoStream):
        stream_cc = stream.codec_context
        cc = stream.codec.create("video")

        cc.width = stream.width
        cc.height = stream.height

        cc.pix_fmt = stream_cc.pix_fmt
        cc.bit_rate = stream_cc.bit_rate
        cc.time_base = stream_cc.time_base
        cc.color_primaries = stream_cc.color_primaries
        cc.color_range = stream_cc.color_range
        cc.color_trc = stream_cc.color_trc

        cc.framerate = stream_cc.framerate
        cc.gop_size = stream_cc.gop_size
        cc.qmax = stream_cc.qmax
        cc.qmin = stream_cc.qmin

        cc.options = stream_cc.options

        return cc

    def _encode_video_frame(self, stream: av.VideoStream, frame: av.VideoFrame | None):
        cc = self._codex_contexts.get(stream.index)
        if frame is None:
            if cc is not None:
                for packet in cc.encode_lazy():
                    packet.stream = stream
                    yield packet
            return

        if cc is None:
            stream.width = frame.width
            stream.height = frame.height
            cc = stream.codec_context
            if cc.coded_width == cc.coded_height == 0:
                pass
            elif (cc.coded_width, cc.coded_height) == (frame.width, frame.height):
                pass
            else:
                cc = self.create_codec_context(stream)

            self._codex_contexts[stream.index] = cc
        elif (stream.width, stream.height) != (frame.width, frame.height):
            stream.width = frame.width
            stream.height = frame.height
            for packet in cc.encode_lazy():
                packet.stream = stream
                yield packet
            cc = self.create_codec_context(stream)
            self._codex_contexts[stream.index] = cc

        for packet in cc.encode_lazy(frame):
            packet.stream = stream
            yield packet

    def encode_video_frame(self, frame: av.VideoFrame):
        frames = [frame]
        if self.alpha_extractor is not None:
            frames.append(self.alpha_extractor(frame))

        for stream, frame in zip(self.streams, frames):
            frame.time_base = TIME_BASE
            frame.pts = round(self.__frames / self.fps / TIME_BASE)
            yield from self._encode_video_frame(stream, frame)

        self.__frames += 1

    def write_video_frame(self, frame: av.VideoFrame):
        for packet in self.encode_video_frame(frame):
            self.container.mux_one(packet)

    def encode_video_frames(self, iterator: Iterable[av.VideoFrame]):
        for frame in iterator:
            for packet in self.encode_video_frame(frame):
                yield packet

    def write_audio(self, audio_segment: AudioSegment):
        audio_segment = (
            audio_segment.set_channels(self.audio_stream.layout.nb_channels)
            .set_sample_width(self.audio_stream.format.bytes)
            .set_frame_rate(self.audio_stream.sample_rate)
        )
        frame = av.AudioFrame.from_ndarray(
            np.array(audio_segment.get_array_of_samples()).reshape(1, -1),
            format=self.audio_stream.format.name,
            layout=self.audio_stream.layout.name,
        )
        frame.sample_rate = audio_segment.frame_rate
        self.write_audio_frame(frame)

    def write_audio_frame(self, frame: av.AudioFrame):
        for packet in self.encode_audio_frames([frame]):
            self.container.mux_one(packet)

    def encode_audio_frames(self, iterator: Iterable[av.AudioFrame]):
        for frame in iterator:
            for packet in self.audio_stream.codec_context.encode_lazy(frame):
                packet.stream = self.audio_stream
                yield packet

    def flush(self):
        if self.future is not None:
            self.future.result()
            del self.future
            self.future = None

        for stream in self.streams:
            for packet in self._encode_video_frame(stream, None):
                self.container.mux_one(packet)

        if self.audio_stream is not None:
            self.container.mux(self.audio_stream.encode())

        if self.alpha_extractor is not None:
            self.alpha_extractor.close()

    def __exit__(
        self,
        t: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ):
        if exc is None:
            self.flush()
        super().__exit__(t, exc, tb)
